-----------------------------------------------------------------------------
-- INTEL DEVELOPER'S SOFTWARE LICENSE AGREEMENT
--
-- BY USING THIS SOFTWARE, YOU ARE AGREEING TO BE BOUND
-- BY THE TERMS OF THIS AGREEMENT.  DO NOT USE THE SOFTWARE
-- UNTIL YOU HAVE CAREFULLY READ AND AGREED TO THE FOLLOWING
-- TERMS AND CONDITIONS.  IF YOU DO NOT AGREE TO THE TERMS
-- OF THIS AGREEMENT, PROMPTLY RETURN THE SOFTWARE PACKAGE
-- AND ANY ACCOMPANYING ITEMS.
--
-- IF YOU USE THIS SOFTWARE, YOU WILL BE BOUND BY THE TERMS
-- OF THIS AGREEMENT.
--
-- LICENSE:  Intel Corporation ("Intel") grants you
-- the non-exclusive right to use the enclosed software
-- program ("Software").  You will not use, copy, modify,
-- display, rent, sell or transfer the Software or any portion
-- thereof, except as provided in this Agreement.
--
-- System OEM Developers may:
-- 1.      copy the Software for internal support, backup
--         or archival purposes;
-- 2.      internally install, use, display, or distribute
--         Intel owned Software in object code format;
-- 3.      internally modify Software source code that
--         Intel makes available to you for internal use
--         only as an OEM Developer;
-- 4.      internally install, use, display, modify, distribute,
--         and/or make derivatives ("Derivatives") of Intel owned
--         Software ONLY if you are a System OEM Developer and
--         NOT an end-user.
--
-- RESTRICTIONS:
--
-- YOU WILL NOT:
-- 1.      copy the Software, in whole or in part, except as
--         provided for in this Agreement;
-- 2.      decompile or reverse engineer any Software provided
--         in object code format;
-- 3.      distribute any Software or Derivative code to any
--         end-users, unless approved by Intel in a prior writing.
--
-- TRANSFER: You may not transfer the Software to any third
-- party without Intels prior written consent.
--
-- OWNERSHIP AND COPYRIGHT OF SOFTWARE: Title to the Software
-- and all copies thereof remain with Intel or its vendors.
-- The Software is copyrighted and is protected by United States
-- and international copyright laws.  You will not remove the
-- copyright notice from the Software.  You agree to prevent
-- any unauthorized copying of the Software.
--
-- DERIVATIVE WORK: OEM Developers that make Derivatives will
-- not be required to provide Intel with a copy of the source
-- or object code.  Any modification of Software shall be at
-- your sole risk and expense. No Software or Derivative
-- distribution to any third party is permitted under this
-- Agreement.
--
-- DUAL MEDIA SOFTWARE: If the Software package contains
-- multiple media, you may only use the medium appropriate
-- for your system.
--
-- WARRANTY: Intel warrants that it has the right to license
-- you to use, modify, display, or distribute the Software as
-- provided in this Agreement. The Software is provided "AS IS"
-- without WARRANTY of any kind.  Intel makes no representations
-- to upgrade, maintain, or support the Software at any time.
--
--
-- THE ABOVE WARRANTIES ARE THE ONLY WARRANTIES OF ANY
-- KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WARRANTIES
-- OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE
-- OR INFRINGEMENT OF ANY PATENT, COPYRIGHT OR OTHER
-- INTELLECTUAL PROPERTY RIGHT.
--
-- LIMITATION OF LIABILITY: NEITHER INTEL NOR ITS
-- VENDORS OR AGENTS SHALL BE LIABLE FOR ANY LOSS
-- OF PROFITS, LOSS OF USE, LOSS OF DATA, INTERRUPTION
-- OF BUSINESS, NOR FOR INDIRECT, SPECIAL, INCIDENTAL
-- OR CONSEQUENTIAL DAMAGES OF ANY KIND WHETHER UNDER
-- THIS AGREEMENT OR OTHERWISE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGES.
--
-- TERMINATION OF THIS LICENSE: Intel reserves the right
-- to conduct or have conducted audits to verify your
-- compliance with this Agreement.  Intel may terminate
-- this Agreement at any time if you are in breach of
-- any of its terms and conditions.  Upon termination,
-- you will immediately destroy, and certify in writing
-- the destruction of, the Software or return all copies
-- of the Software and documentation to Intel.
--
-- U.S. GOVERNMENT RESTRICTED RIGHTS: The Software and
-- documentation were developed at private expense and
-- are provided with "RESTRICTED RIGHTS".  Use, duplication
-- or disclosure by the Government is subject to restrictions
-- as set forth in FAR52.227-14 and DFAR252.227-7013 et seq.
-- or its successor.
--
-- EXPORT LAWS: You agree that the distribution and
-- export/re-export of the Software is in compliance
-- with the laws, regulations, orders or other restrictions
-- of the U.S. Export Administration Regulations.
--
-- APPLICABLE LAW: This Agreement is governed by the
-- laws of the State of California and the United States,
-- including patent and copyright laws.  Any claim
-- arising out of this Agreement will be brought in
-- Santa Clara County, California.
--
-- Copyright 1996, Intel Corporation, All Rights Reserved
--
--   File name :  i28f008s3_bfm_arc.vhd
--   Title     :  []
--   Module    :  []
--
--   Purpose   :  []
--
--   Roadmap   :
-------------------------------------------------------------------------------
--   Modification History :
--          Date                Author          Revision        Comments
-- Fri Mar  3 09:06:27 1995     Rupinder Hundal Rev2.0_0.0      Creation
-- Tue Mar 21 1995              RH              Rev2.0_0.1      process added
-- Mon Mar 27 1995              RH              Rev2.0_0.2      status process added
-- Fri Mar 31 1995              RH              Rev2.0_0.3      control pin logic process
-- Tue Apr 04 1995              RH              Rev2.0_0.4      array process modified
-- Wed Apr 12 1995              RH              Rev2.0_0.5      face process added
-- Thu Apr 13 1995              RH              Rev2.0_0.6      timer processes added
-- Fri/Mon Apr 14/17 1995       RH              Rev2.0_0.7      debug edits 
-- Wed Jul 19 1995              RH              Rev2.0_0.9      add comments
-- Wed Oct 04 1995              RH              Rev2.0_1.0      device id -> A6
-- Fri Oct  6 15:09:20 1995     RH              Rev2.0_1.1      
--              edit to output status during suspend states, setup states and erase botch.
-- Mon Apr  1 11:24:08 1996     RH              Rev3.0_0.0
--              FACE: HV override implemented for LockBlock, Lockmaster, Program & Erase
--              Timing Checker: tPHHWH & tPHHEH timing checks added.
-- Tue Apr  2 17:05:16 1996     RH
--              changed tProgram,tErase,tLockProgram & tLockErase from constants to generics.
--              suspend latencies implemented
-- Thu Apr  4 11:42:20 1996     RH
-------------------------------------------------------------------------------

ARCHITECTURE rev1 OF i28f008s3_bfm IS

--  constants
  CONSTANT      vend_id          : std_ulogic_vector(7 DOWNTO 0) := "10001001"; --  vendor id = 89H
  CONSTANT      dev_id           : std_ulogic_vector(7 DOWNTO 0) := "10100110"; --  device id = A6H for 8M-SC
  CONSTANT      ready_low        : TIME := 200 ns;
  CONSTANT      tClearStatus     : TIME := 10 ns;
  CONSTANT      Vlko             : std_uV := 3 V;                               -- Vcc program/erase lock out voltage
  CONSTANT      Vpplk            : std_uV := 2 V;                               -- Vpp program/erase lock out voltage

--  types
TYPE State_Machine_Type IS (ReadArray, ReadStatus, ReadID, LockBlock, LockBlockAct,
			    ProgramSetup, ProgramAct, ProgramSuspend, ProgramSuspToArray,
			    EraseSetup, EraseBotched, EraseAct, EraseSuspend,
			    EraseSuspToArray, ClearStatus);     -- only 14 are csm states

TYPE Control_Machine_Type IS (Read, Program, Erase, Lock, Idle);

--  signals
  SIGNAL        addr_buf                : std_ulogic_vector(mem_size DOWNTO 0); --  address related paths
  SIGNAL        csm_addr                : std_ulogic_vector(mem_size DOWNTO 0);
  SIGNAL        mem_addr                : std_ulogic_vector(mem_size DOWNTO 0);
  
  SIGNAL        data_buf                : bit_vector(7 DOWNTO 0);               --   data related paths
  SIGNAL        csm_data                : std_ulogic_vector(7 DOWNTO 0);
  SIGNAL        mem_data                : std_ulogic_vector(7 DOWNTO 0);
  SIGNAL        mem_data_out            : std_ulogic_vector(7 DOWNTO 0);
  SIGNAL        mux_data_out            : std_ulogic_vector(7 DOWNTO 0);

  SIGNAL        output_en               : std_ulogic;                           --   control pin logic outputs
  SIGNAL        output_xs               : std_ulogic; 
  SIGNAL        addr_en                 : std_ulogic;
  SIGNAL        data_en                 : std_ulogic;
  SIGNAL        csm_clock               : std_ulogic;
  SIGNAL        reset                   : std_ulogic;

  SIGNAL        id_or_status            : std_ulogic_vector(7 DOWNTO 0);        -- status outputs
  SIGNAL        read_lock               : std_ulogic;
  SIGNAL        read_array              : std_ulogic;

  SIGNAL        csm_mux                 : std_ulogic;                           --  CSM outputs
  SIGNAL        command                 : State_Machine_Type;

  SIGNAL        control_status          : std_ulogic_vector(7 DOWNTO 0);        --  FACE outputs
  SIGNAL        control_rdy_idle        : std_ulogic_vector(2 DOWNTO 0);
  SIGNAL        control_command         : Control_Machine_Type;
  SIGNAL        ready                   : std_logic := '1';                     -- used within FACE

  
  SIGNAL        cur_state               : State_Machine_Type;                   --  used within CSM
  SIGNAL        block_status_sig        : bit_vector(31 DOWNTO 0);              -- array output, face uses

  SIGNAL        Lock_Program_Timer_Start : bit;                                 --  Timer control signals
  SIGNAL        Lock_Erase_Timer_Start   : bit;
  SIGNAL        Program_Timer_start      : bit;
  SIGNAL        Erase_Timer_Start        : bit;
  SIGNAL        Lock_Program_Done        : bit;
  SIGNAL        Lock_Erase_Done          : bit;
  SIGNAL        Program_Done             : bit;
  SIGNAL        Erase_Done               : bit;

  SIGNAL        output_en_1              : std_ulogic;                          -- read timing signals
  SIGNAL        output_en_2              : std_ulogic;
  SIGNAL        output_en_3              : std_ulogic;
  SIGNAL        output_en_4              : std_ulogic;
  SIGNAL        output_xs_2              : std_ulogic;
  SIGNAL        output_xs_3              : std_ulogic;
  SIGNAL        output_xs_4              : std_ulogic;
  
BEGIN  --  rev1 
  

-------------------------------------------------------------------------------
--  Process:   control_pin_logic
--  Purpose:
--  Inputs:    rpb, ceb, web, oeb
--  Outputs:  data_en, addr_en, output_en, output_xs, csm_clock, reset
-------------------------------------------------------------------------------
control_pin_logic : PROCESS (rpb, ceb, web, oeb, addr,
			     output_en_1, output_en_2, output_en_3, output_en_4,
			     output_xs_2, output_xs_3, output_xs_4)
  
BEGIN  --  PROCESS control_pin_logic 

  IF (rpb = '1') OR (rpb = 'H') THEN                            -- device is not in powerdown
    reset <= '0';                                               -- do not reset the device
    
    IF (ceb ='0') THEN                                          -- addr_en = (rpb)(/ceb)
      addr_en <= '1';
    ELSE
      addr_en <= '0';
    END IF;

    IF (ceb = '0' AND web = '0' AND oeb ='1') THEN              -- csm_clock = (/rpb)(/ceb)(/web)(oeb)
      csm_clock <= '1';                                         -- used to clock the csm
      data_en <= '1';                                           -- data_en signal is equiv, could delete
    ELSE                                                        -- used to enable the data buffers
      csm_clock <= '0';
      data_en <= '0';
    END IF;

    IF (command = ReadArray) OR (command = EraseSuspToArray) OR
      (command = ProgramSuspToArray) OR (command = ClearStatus) OR
      (command = ReadID) OR (command = ReadStatus) OR (command = ProgramAct) OR
      (command = EraseAct) OR (command = LockBlockAct) OR
      (command = EraseSuspend) OR (command = ProgramSuspend) OR
      (command = LockBlock) OR (command = ProgramSetup) OR
      (command = EraseSetup) OR (command = EraseBotched) THEN   -- output the array/id/status in these states, 
								-- so use the read timings.
      IF (rpb'EVENT) AND (rpb = '1') THEN                       -- delay after rising edge of rpb 
	output_en_1 <= '1' After tPHQV;
      END IF;

      IF (ceb'EVENT) AND (ceb = '0') THEN                       -- delay after falling edge of ceb 
	output_xs_2 <= '1' After tELQX, '0' After tELQV - tELQX;
	output_en_2 <= '1' After tELQV;
      END IF;

      IF (oeb'EVENT) AND (oeb = '0') THEN                       -- delay after falling edge of oeb
	output_xs_3 <= '1' After tGLQX, '0' After tGLQV - tGLQX;
	output_en_3 <= '1' AFTER tGLQV;
      END IF;

      IF (addr'EVENT) THEN                                      -- delay after address transition
	output_xs_4 <= '1' After tOH, '0' After tAVQV; 
	output_en_4 <= '0' After tOH, '1' After tAVQV;
      END IF;

      IF (ceb'EVENT) AND (ceb = '1') THEN                       -- delay after rising edge of ceb
	output_xs_2 <= '1' After tOH, '0' After tEHQZ - tOH;
	output_en_2 <= '0' After tOH;
      END IF;

      IF (oeb'EVENT) AND (oeb = '1') THEN                       -- delay after rising edge of oeb
	output_xs_3 <= '1' After tOH, '0' After tGHQZ - tOH;
	output_en_3 <= '0' After tOH;
      END IF;

      IF (web'EVENT) AND (web = '0') THEN                       -- reseting all output timings during a write,
	output_xs_2 <= '0';                                     -- web falling edge
	output_xs_3 <= '0';
	output_xs_4 <= '0';
	output_en_2 <= '0';
	output_en_3 <= '0';
	output_en_4 <= '0';
      END IF;

      IF (web'EVENT) AND (web = '1') AND (ceb = '0') THEN       -- if ceb is already low during a web rising edge,
	output_en_2 <= '1';                                     -- set output_en to 1 immediately.
      END IF;

      IF (output_en_1'EVENT) OR (output_en_2'EVENT) OR (output_en_3'EVENT) OR
	(output_en_4'EVENT) OR (output_xs_2'EVENT) OR (output_xs_3'EVENT) OR
	(output_xs_4'EVENT) THEN                                -- if any of the intermediate output_en or output_xs
								-- has an event, then recalculate output_en and 
								-- output_xs signals going to the output buffer.
	IF (command = ReadStatus) OR (command = ProgramAct) OR (command = EraseAct) OR
	  (command = LockBlockAct) OR (command = EraseSuspend) OR 
	  (command = ProgramSuspend) OR (command = LockBlock) OR 
	  (command = ProgramSetup) OR (command = EraseSetup) OR 
	  (command = EraseBotched) THEN                         -- during these csm states the timings due to
								-- address transitions are irrelevant.
	  IF (output_en_1 = '1') AND (output_en_2 = '1') AND (output_en_3 = '1') THEN
	    output_en <= '1';
	  ELSE
	    output_en <= '0';
	  END IF;

	ELSE                                                    -- during the other states all 4 output_en signals
								-- must be enabled for the output buffer to be enabled.
	  IF (output_en_1 = '1') AND (output_en_2 = '1') AND (output_en_3 = '1') AND
	    (output_en_4 = '1') THEN
	    output_en <= '1';
	  ELSE
	    output_en <= '0';
	  END IF;

	END IF;

	IF (output_xs_2 = '1') OR (output_xs_3 = '1') OR (output_xs_4 = '1') THEN       -- if any of the output_xs is set
	  output_xs <= '1';                                     -- the output buffer points to X's.
	ELSE
	  output_xs <= '0';
	END IF;
	
      END IF;
    END IF;                                                     -- end of the read timings

    IF (ceb = '0' AND oeb = '0' AND web = '0') THEN             -- oeb and web cannot both be low
      ASSERT (NOT verbose)
	REPORT "control_pin_logic: oeb and web cannot both be low."
      SEVERITY error;
    END IF;

  ELSE                                                          -- device is in powerdown(rpb = '0')
    reset <= '1';                                               -- reset the device
    output_en_1 <= '0';                                         -- rpb based output_en is reset 
    output_en_4 <= '0';                                         -- address transition based output_en is reset 
    output_en <= '0';                                           -- output buffer signals are reset
    output_xs <= '0';

  END IF;

  
END PROCESS control_pin_logic;

-------------------------------------------------------------------------------
--  Address/Data Buffers
-------------------------------------------------------------------------------

addr_buf <= addr WHEN addr_en = '1' ELSE                        -- address on the pads is allowed through the buffer 
	    addr_buf;

data_buf <= To_bitvector(data) WHEN data_en = '1' ELSE          -- data on the pads is allowed through the buffer
	    data_buf;

-------------------------------------------------------------------------------
--  Address/Data Mux
-------------------------------------------------------------------------------

mem_addr <= csm_addr WHEN csm_mux = '1' ELSE                    -- either the address from the csm or the pads goes 
	    addr_buf;                                           -- to the array.

mem_data <= csm_data WHEN csm_mux = '1' ELSE                    -- either the data from the csm or X's are driven 
	    "XXXXXXXX";                                         -- to the array.

-------------------------------------------------------------------------------
--  Process:   csm
--  Purpose:
--  Inputs:    control_rdy_idle,addr_buf,data_buf,csm_clock,reset
--  Outputs:   command,command_go,csm_mux,csm_data,csm_addr,clr_status
-------------------------------------------------------------------------------

csm : PROCESS (csm_clock, reset, cur_state)                     -- only clock, reset & change in current state wakeup the CSM
	
	VARIABLE        nxt_state                : State_Machine_Type := ReadArray;
	VARIABLE        latch_rdy_idle           : std_ulogic_vector(2 DOWNTO 0);
	VARIABLE        control_rdy,e_idle,p_idle: BOOLEAN := FALSE;
	VARIABLE        idle_state               : BOOLEAN := FALSE;
	VARIABLE        Face_Act                 : BOOLEAN := FALSE;
	
BEGIN  --  PROCESS csm 

  IF ((csm_clock'EVENT) AND (csm_clock = '1') AND (csm_clock'LAST_VALUE = '0')) THEN    -- at the rising edge of clock(web falling)
    latch_rdy_idle := control_rdy_idle;                         -- latch the ready and idle status from FACE
  END IF;

  IF (reset = '1') THEN                                         -- when device is reset (rpb = 0)
    cur_state <= ReadArray;                                     -- it goes into ReadArray mode
    latch_rdy_idle := control_rdy_idle;                         -- the cleared/reset status from FACE is latched
  END IF;

  IF ((csm_clock'EVENT) AND (csm_clock = '0') AND (csm_clock'LAST_VALUE = '1')) THEN    -- at the falling edge of clock(web rising)

    IF (latch_rdy_idle(2) = '1') THEN                           -- convert the 3-bit vector to individual boolean signals
      control_rdy := TRUE;                                      -- for easier if-then-else structure in the csm
    ELSE
      control_rdy := FALSE;
    END IF;
    IF (latch_rdy_idle(1) = '1') THEN
      e_idle := TRUE;
    ELSE
      e_idle := FALSE;
    END IF;
    IF (latch_rdy_idle(0) = '1') THEN
      p_idle := TRUE;
    ELSE
      p_idle := FALSE;
    END IF;


  CASE cur_state IS                                             -- next state based on the current state and user write
    
    WHEN ReadArray =>
      idle_state := TRUE;                                       -- all user commands are available

    WHEN ReadStatus =>
      idle_state := TRUE;                                       -- all user commands are available
      
    WHEN ReadID =>
      idle_state := TRUE;                                       -- all user commands are available
      
    WHEN LockBlock =>
      nxt_state := LockBlockAct;                                -- next state must be LockBlockAct
      Face_Act := TRUE;                                         -- first of a 2 cycle command

    WHEN LockBlockAct =>
      IF (NOT control_rdy) THEN                                 -- !ready
	nxt_state := LockBlockAct;
      ELSIF (control_rdy) THEN                                  -- ready
	idle_state := TRUE;
      END IF;

    WHEN ProgramSetup =>
      nxt_state := ProgramAct;                                  -- next state must be ProgramAct
      Face_Act := TRUE;                                         -- first of a 2 cycle command

    WHEN ProgramAct =>
      IF (NOT control_rdy) THEN                                 --   !ready
	
	 IF (data_buf = X"B0") THEN                             --   Suspend
	   nxt_state := ProgramSuspend;
	 ELSE                                                   --   any other command
	   nxt_state := ProgramAct;
	 END IF;
	 
      ELSIF (control_rdy) THEN                                  --   ready
	
	IF (e_idle) THEN                                        --   erase suspended
	  IF ((data_buf = X"10" OR data_buf = X"40")) THEN      --   Program
	    nxt_state := ProgramSetup;
	  ELSIF (data_buf = X"D0") THEN                         --   Resume(erase)
	    nxt_state := EraseAct;
	  ELSIF (data_buf = X"70") THEN                         --   Read status
	    nxt_state := EraseSuspend;
	  ELSE                                                  --   any other command
	    nxt_state := EraseSuspToArray;
	  END IF;
	ELSIF (NOT e_idle) THEN                                 --   !erase suspended
	  idle_state := TRUE;
	END IF;

      END IF;

    WHEN ProgramSuspend =>

      IF (NOT control_rdy) THEN                                 --   !ready
	
	IF (data_buf = X"D0") THEN                              --   Resume
	  nxt_state := ProgramAct;
	ELSE                                                    --   !Resume
	  nxt_state := ProgramSuspend;
	END IF;
	
      ELSIF (control_rdy) THEN                                  --   ready
	
	IF (p_idle) THEN                                        --   program suspended successfully
	  
	  IF (data_buf = X"D0") THEN                            --   Resume
	    nxt_state := ProgramAct;
	  ELSIF (data_buf = X"70") THEN                         --   Read Status
	    nxt_state := ProgramSuspend;
	  ELSE                                                  --  any other command
	    nxt_state := ProgramSuspToArray;
	  END IF;
	  
	ELSIF (NOT p_idle) THEN                                 --   program already completed, not suspended
	  
	  IF (e_idle) THEN                                      --   erase suspended
	    
	    IF (data_buf = X"10" OR data_buf = X"40") THEN      --   Program
	      nxt_state := ProgramSetup;
	    ELSIF (data_buf = X"D0") THEN                       --   Resume(erase)
	      nxt_state := EraseAct;
	    ELSIF (data_buf = X"70") THEN                       --   Read Status
	      nxt_state := EraseSuspend;
	    ELSE                                                --   any other command
	      nxt_state := EraseSuspToArray;
	    END IF;
	    
	  ELSIF (NOT e_idle) THEN                               --   !erase suspended
	    idle_state := TRUE;
	    
	  END IF;
	END IF;
      END IF;

    WHEN ProgramSuspToArray =>

      IF (data_buf = X"70") THEN                                --   Read Status
	nxt_state := ProgramSuspend;
      ELSIF (data_buf = X"D0") THEN                             --   Resume
	nxt_state := ProgramAct;
      ELSE                                                      --   any other command
	nxt_state := ProgramSuspToArray;
      END IF;

    WHEN EraseSetup =>

      IF (data_buf = X"D0") THEN                                --   Erase Confirm
	nxt_state := EraseAct;                                  -- next state must be EraseAct
	Face_Act := TRUE;                                       -- first of a 2 cycle command
      ELSE                                                      --   any other command
	nxt_state := EraseBotched;
      END IF;

    WHEN EraseBotched =>
      idle_state := TRUE;                                       -- all user commands are available

    WHEN EraseAct =>

      IF (NOT control_rdy) THEN                                 --   !ready

	IF (data_buf = X"B0") THEN                              --   Suspend
	  nxt_state := EraseSuspend;
	ELSE                                                    --   !Suspend
	  nxt_state := EraseAct;
	END IF;

      ELSE                                                      --   ready
	idle_state := TRUE;                                     -- all user commands are available

      END IF;

    WHEN EraseSuspend =>

      IF (NOT control_rdy) THEN                                 --   !ready (not suspended yet)
	
	IF (data_buf = X"D0") THEN                              --   Resume
	  nxt_state := EraseAct;
	ELSE                                                    --   !Resume
	  nxt_state := EraseSuspend;
	END IF;
	
      ELSIF (control_rdy) THEN                                  --   ready

	IF (e_idle) THEN                                        --   erase suspended
	    
	  IF (data_buf = X"10" OR data_buf = X"40") THEN        --   program
	    nxt_state := ProgramSetup;
	  ELSIF (data_buf = X"D0") THEN                         --   Resume(erase)
	    nxt_state := EraseAct;
	  ELSIF (data_buf = X"70") THEN                         --   Read Status
	    nxt_state := EraseSuspend;
	  ELSE                                                  --   any other command
	    nxt_state := EraseSuspToArray;
	  END IF;
	    
	ELSIF (NOT e_idle) THEN                                 --   erase already completed, not suspended
	    idle_state := TRUE;

	END IF;
      END IF;

    WHEN EraseSuspToArray =>

      IF (data_buf = X"70") THEN                                --   Read Status
	nxt_state := EraseSuspend;
      ELSIF (data_buf = X"D0") THEN                             --   Resume
	nxt_state := EraseAct;
      ELSIF (data_buf = X"10" OR data_buf = X"40") THEN         --   Program
	nxt_state := ProgramSetup;
      ELSE                                                      --   any other command
	nxt_state := EraseSuspToArray;
      END IF;

    WHEN OTHERS =>                                              --   this should never happen

      ASSERT (NOT verbose)
	REPORT "State Machine is Lost!!"
      SEVERITY ERROR;
      nxt_state := ReadArray;

  END CASE;                                                     --  end case cur_state

  IF (idle_state) THEN                                          -- if this flag is set 
    
    idle_state := FALSE;                                        -- reset the flag
    
    IF (data_buf = X"10" OR data_buf = X"40") THEN              -- all user commands are available
      nxt_state := ProgramSetup;
    ELSIF (data_buf = X"20") THEN
      nxt_state := EraseSetup;
    ELSIF (data_buf = X"70") THEN
      nxt_state := ReadStatus;
    ELSIF (data_buf = X"50") THEN
      nxt_state := ClearStatus;
    ELSIF (data_buf = X"90") THEN
      nxt_state := ReadID;
    ELSIF (data_buf = X"60") THEN
      nxt_state := LockBlock;
    ELSE
      nxt_state := ReadArray;
    END IF;

  END IF;


    IF (Face_Act) THEN                                          -- if a 2 cycle command(program,erase,lock)
      Face_Act := FALSE;                                        -- reset the flag
      csm_addr <= addr_buf;                                     -- latch the current address - signal csm_addr
      csm_data <= To_StdULogicVector(data_buf);                 -- latch the current data
      csm_mux <= '1';                                           -- the addr/data muxes should point to csm addr/data
    ELSE                                                        -- if not a 2 cycle command
      csm_mux <= '0';
    END IF;

    IF (nxt_state = ClearStatus) THEN                           -- ClearStatus is not a csm 'state' so,
      command <= ClearStatus,                                   -- command is ClearStatus for a short time,
		 ReadArray AFTER tClearStatus;                  -- then returns to the ReadArray state
      cur_state <= ReadArray AFTER tClearStatus;                -- signal cur_state, used to sequence through csm,is ReadArray
    ELSE                                                        -- if any other nxt_state (besides ClearStatus)
      cur_state <= nxt_state;                                   -- nxt_state is transferred to both signals
      command <= nxt_state;
    END IF;

  END IF;                                                       -- end of everything done at the rising edge of clock(web falling)

END PROCESS csm;


-------------------------------------------------------------------------------
--  Process:   face
--  Purpose:
--  Inputs: command, reset
--  Outputs:    control_command, control_status, control_rdy_idle 
-------------------------------------------------------------------------------
face : PROCESS(command, reset, ready,
	 Lock_Program_Done, Lock_Erase_Done, Program_Done, Erase_Done)
	 
  VARIABLE      e_susp, e_fail, p_fail, vpp_low, p_susp, l_fail  : std_logic := '0';    -- suspend/fail status bits
  VARIABLE      erasing_blk      : integer := -1;               -- to store the block being erased
  VARIABLE      programming_blk  : integer := -1;               -- to store the block being programmed
  VARIABLE      master_lock      : bit := '0';                  -- master lock status
  
BEGIN  --  PROCESS face 

  CASE command IS
    WHEN ReadArray =>
      control_command <= Read;                                  -- array process should Read

    WHEN ReadStatus =>
      control_command <= Idle;                                  -- array process should do nothing

    WHEN ClearStatus =>
      control_command <= Idle;                                  -- array process should do nothing
      ready <= '1';                                             -- reset all status bits
      e_susp := '0';
      e_fail := '0';
      p_fail := '0';
      vpp_low := '0';
      p_susp := '0';
      l_fail := '0';
      
    WHEN ReadID =>
      control_command <= Idle;                                  -- array process should do nothing

    WHEN LockBlock =>
      control_command <= Idle;                                  -- array process should do nothing

    WHEN LockBlockAct =>
      IF (command'EVENT) THEN                                   -- if command has had an event

	IF (csm_data = "11110001") AND (rpb = 'H') THEN         --  (data = X"F1")
	  IF ((vccpad >= Vlko) AND (vpppad > Vpplk)) THEN       -- not vcc lock out voltage or vpp lock out voltage
	    control_command <= Lock;                            -- array process should Lock the master lock
	    Lock_Program_Timer_Start <= '1' AFTER tWHRL;                --  program time for locking
	    ready <= '0' AFTER tWHRL;                           -- rd_byb goes low after tWHRL
	    master_lock := '1';                                 -- variable kept here so that array_proc doesn't need to 
								-- pass a signal for master lock status.
	  ELSIF (vpppad <= Vpplk) THEN                          -- vpp lock out voltage
	    p_fail := '1';
	    vpp_low := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  ELSIF (vccpad < Vlko) THEN                            -- vcc lock out voltage
	    p_fail := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  END IF;

	ELSIF (csm_data = "00000001") AND ((master_lock = '0')  --  (data = X"01") and master lock is not set
		OR (master_lock = '1' AND rpb = 'H')) THEN      -- or master lock is set but there is HV override
	  IF ((vccpad >= Vlko) AND (vpppad > Vpplk)) THEN       -- not vcc lock out voltage or vpp lock out voltage
	    control_command <= Lock;                            -- array process should Lock the block based on addr
	    ready <= '0' AFTER tWHRL;                           -- delay before rd_byb goes low
	    Lock_Program_Timer_Start <= '1' AFTER tWHRL;        --  program time for locking
	  ELSIF (vpppad <= Vpplk) THEN                          -- vpp lock out voltage
	    p_fail := '1';
	    vpp_low := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  ELSIF (vccpad < Vlko) THEN                            -- vcc lock out voltage
	    p_fail := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  END IF;

	ELSIF (csm_data = "11010000") AND ((master_lock = '0')  --  (data = X"D0") and master lock is not set
		OR (master_lock = '1' AND rpb = 'H')) THEN      -- or master lock is set but there is HV override
	  IF ((vccpad >= Vlko) AND (vpppad > Vpplk)) THEN       -- not vcc lock out voltage or vpp lock out voltage
	    control_command <= Lock;                            -- array process should unlock blocks
	    ready <= '0' AFTER tWHRL;                           -- delay before rd_byb goes low
	    Lock_Erase_Timer_Start <= '1' AFTER tWHRL;          --  erase time for erasing locks
	  ELSIF (vpppad <= Vpplk) THEN                          -- vpp lock out voltage
	    e_fail := '1';
	    vpp_low := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  ELSIF (vccpad < Vlko) THEN                            -- vcc lock out voltage
	    e_fail := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  END IF;

	ELSIF (csm_data = "11110001") AND (rpb /= 'H') THEN     --  (data = X"F1") and there is no HV override
	  control_command <= Idle;                              -- array process should do nothing
	  p_fail := '1';                                        -- program fail & lock fail bits set
	  l_fail := '1';
	  ready <= '0' AFTER tWHRL,                             -- rd_byb goes low for a short time (ready_low)
		   '1' AFTER tWHRL + ready_low;

	ELSIF (csm_data = "00000001") AND (master_lock = '1')   --  (data = X"01") & master lock is set
		AND (rpb /= 'H') THEN                           -- and there is no HV override
	  control_command <= Idle;                              -- array process should do nothing
	  p_fail := '1';                                        -- program fail & lock fail bits set
	  l_fail := '1';
	  ready <= '0' AFTER tWHRL,                             -- rd_byb goes low for a short time (ready_low)
		   '1' AFTER tWHRL + ready_low;

	ELSIF (csm_data = "11010000") AND (master_lock = '1')   --  (data = X"D0") & master lock is set
		AND (rpb /= 'H') THEN                           -- and there is no HV override
	  control_command <= Idle;                              -- array process should do nothing
	  e_fail := '1';                                        -- erase fail & lock fail bits set
	  l_fail := '1';
	  ready <= '0' AFTER tWHRL,                             -- rd_byb goes low for a short time (ready_low)
		   '1' AFTER tWHRL + ready_low;
	ELSE                                                    -- if incorrect lock command sequence given 
	  e_fail := '1';                                        -- both e_fail & p_fail bits set just like botched but not l_fail
	  p_fail := '1';                                        --    after ??? 
	  ready <= '0' AFTER tWHRL,                             -- rd_byb goes low for a short time (ready_low)
		   '1' AFTER tWHRL + ready_low;
	END IF;
	
      ELSIF (Lock_Program_Done'EVENT) AND (Lock_Program_Done = '1') THEN        -- if program timer is done
	Lock_Program_Timer_Start <= '0';                        -- timer turned off and rd_byb goes high
	ready <= '1';

      ELSIF (Lock_Erase_Done'EVENT) AND (Lock_Erase_Done = '1') THEN            -- if erase timer is done
	Lock_Erase_Timer_Start <= '0';                          -- timer turned off and rd_byb goes high
	ready <= '1';
      END IF;
	
    WHEN ProgramSetup =>
      control_command <= Idle;                                  -- array process should do nothing

    WHEN ProgramAct =>
      IF (command'EVENT) THEN                                   -- if command has had an event

      programming_blk := BitVector_to_Int(To_BitVector(mem_addr(mem_size DOWNTO 16)));  -- store block number being programmed
      IF (block_status_sig(programming_blk) = '0')              -- if block lock bit is not set
	OR (block_status_sig(programming_blk) = '1' AND rpb = 'H') THEN -- or block lock bit is set but there is HV override

      IF (e_susp = '1') THEN                                    -- if erase suspended 
	IF (programming_blk /= erasing_blk) THEN                -- & not attempting to program in the same block being erased
	  IF (p_susp = '0') THEN                                -- & it is a first time program
	    IF ((vccpad >= Vlko) AND (vpppad > Vpplk)) THEN     -- not vcc lock out voltage or vpp lock out voltage
	      control_command <= Program;                       -- array process should program the byte
	      ready <= '0' AFTER tWHRL;                         -- rd_byb goes low after tWHRL
	      Program_Timer_Start <= '1' AFTER tWHRL;           -- program timer started after rd_byb goes low
	    ELSIF (vpppad <= Vpplk) THEN                        -- vpp lock out voltage
	      p_fail := '1';
	      vpp_low := '1';
	      ready <= '0' AFTER tWHRL,
		       '1' AFTER tWHRL + ready_low;             -- rd_byb goes low for a short time (ready_low)
	    ELSIF (vccpad < Vlko) THEN                          -- vcc lock out voltage
	      p_fail := '1';
	      ready <= '0' AFTER tWHRL,
		       '1' AFTER tWHRL + ready_low;             -- rd_byb goes low for a short time (ready_low)
	    END IF;
	  ELSIF (p_susp = '1') THEN                             -- after resume
	    IF ((vccpad >= Vlko) AND (vpppad > Vpplk)) THEN     -- not vcc lock out voltage or vpp lock out voltage
	      control_command <= Idle;                          -- array process should do nothing (already programmed!)
	      p_susp := '0';                                    -- sometime later p_susp bit is cleared (???)
	      ready <= '0' AFTER tWHRL;                         -- rd_byb goes low after tWHRL
	      Program_Timer_Start <= '1' AFTER tWHRL;           -- program timer is restarted
	    ELSIF (vpppad <= Vpplk) THEN                        -- vpp lock out voltage
	      p_fail := '1';
	      vpp_low := '1';
	      ready <= '0' AFTER tWHRL,
		       '1' AFTER tWHRL + ready_low;             -- rd_byb goes low for a short time (ready_low)
	    ELSIF (vccpad < Vlko) THEN                          -- vcc lock out voltage
	      p_fail := '1';
	      ready <= '0' AFTER tWHRL,
		       '1' AFTER tWHRL + ready_low;             -- rd_byb goes low for a short time (ready_low)
	    END IF;
	  END IF;
	ELSE                                                    -- if attempting to program in the same block being erased
	  ASSERT (NOT verbose)
	    REPORT "FACE: Cannot program in a block being erased."
	  SEVERITY warning;
	  p_fail := '1';                                        -- sometime later set p_fail (???)
	  programming_blk := -1;                                -- not programming in any block (0 to 16), used negative integer
	  ready <= '0' AFTER tWHRL,                             -- rd_byb goes low for a short time (ready_low)
		   '1' AFTER tWHRL + ready_low;
	END IF;
	
      ELSE                                                      -- if erase is not suspended
	IF (p_susp = '0') THEN                                  -- & it is a first time program
	  IF ((vccpad >= Vlko) AND (vpppad > Vpplk)) THEN       -- not vcc lock out voltage or vpp lock out voltage
	    control_command <= Program;                         -- array process should program the byte
	    ready <= '0' AFTER tWHRL;                           -- rd_byb goes low after tWHRL
	    Program_Timer_Start <= '1' AFTER tWHRL;             -- program timer started after rd_byb goes low
	  ELSIF (vpppad <= Vpplk) THEN                          -- vpp lock out voltage
	    p_fail := '1';
	    vpp_low := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  ELSIF (vccpad < Vlko) THEN                            -- vcc lock out voltage
	    p_fail := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  END IF;
	ELSE                                                    -- after resume
	  IF ((vccpad >= Vlko) AND (vpppad > Vpplk)) THEN       -- not vcc lock out voltage or vpp lock out voltage
	    control_command <= Idle;                            -- array process should do nothing (already programmed!)
	    p_susp := '0';                                      -- sometime later p_susp bit is cleared (???) 
	    ready <= '0' AFTER tWHRL;                           -- rd_byb goes low after tWHRL
	    Program_Timer_Start <= '1' AFTER tWHRL;             -- program timer is restarted
	  ELSIF (vpppad <= Vpplk) THEN                          -- vpp lock out voltage
	    p_fail := '1';
	    vpp_low := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  ELSIF (vccpad < Vlko) THEN                            -- vcc lock out voltage
	    p_fail := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  END IF;
	END IF;
	
      END IF;

      ELSE                                                      -- if block lock bit is set & RP# is not at VHH
	p_fail := '1';                                          -- both p_fail & l_fail bits set after ??? 
	l_fail := '1';
	programming_blk := -1;                                  -- not programming in any block (0 to 16), used negative integer 
	ready <= '0' AFTER tWHRL,
		 '1' AFTER tWHRL + ready_low;                   -- rd_byb goes low for a short time (ready_low) 
      END IF;

      ELSIF (Program_Done'EVENT) AND (Program_Done = '1') THEN  -- if program timer is done
	Program_Timer_Start <= '0';                             -- timer turned off and rd_byb goes high
	ready <= '1';
	programming_blk := -1;                                  -- done programming, variable reset

      END IF;

    WHEN ProgramSuspend =>
      control_command <= Idle;                                  -- array process should do nothing
      p_susp := '1';                                            -- sometime later p_susp bit is set (???)
      Program_Timer_Start <= '0';                               -- timer turned off and rd_byb goes high 
      ready <= '1' AFTER tWHRH1;                                -- Program Suspend latency

    WHEN ProgramSuspToArray =>
      control_command <= Read;                                  -- array process should Read

    WHEN EraseSetup =>
      control_command <= Idle;                                  -- array process should do nothing

    WHEN EraseBotched =>
      control_command <= Idle;                                  -- array process should do nothing
      e_fail := '1';                                            -- both e_fail & p_fail bits set after (???)
      p_fail := '1';                                            -- ready remains '1' in botched state

    WHEN EraseAct =>
      IF (command'EVENT) THEN                                   -- if command has had an event

	erasing_blk := BitVector_to_Int(To_BitVector(mem_addr(mem_size DOWNTO 16)));--  store the block number being erased
	IF (block_status_sig(erasing_blk) = '0')                -- if block lock bit is not set
	  OR (block_status_sig(erasing_blk) = '1' AND rpb = 'H') THEN   -- or block lock bit is set but there is HV override

	IF (e_susp = '0') THEN                                  -- & it is a first time erase
	  IF ((vccpad >= vlko) AND (vpppad > Vpplk)) THEN       -- not vcc lock out voltage or vpp lock out voltage
	    control_command <= Erase;                           -- array process should erase the block
	    ready <= '0' AFTER tWHRL;                           -- rd_byb goes low after tWHRL
	    Erase_Timer_Start <= '1' AFTER tWHRL;                       -- erase timer started after rd_byb goes low
	  ELSIF (vpppad <= Vpplk) THEN                          -- vpp lock out voltage
	    e_fail := '1';
	    vpp_low := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  ELSIF (vccpad < Vlko) THEN                            -- vcc lock out voltage
	    e_fail := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  END IF;
	ELSIF (e_susp ='1') THEN                                -- after resume
	  IF ((vccpad >= vlko) AND (vpppad > Vpplk)) THEN       -- not vcc lock out voltage or vpp lock out voltage
	    control_command <= Idle;                            -- array process should do nothing (already erased!)
	    e_susp := '0';                                      -- sometime later e_susp bit is cleared (???) 
	    ready <= '0' AFTER tWHRL;                           -- rd_byb goes low after tWHRL
	    Erase_Timer_Start <= '1' AFTER tWHRL;               -- erase timer is restarted
	  ELSIF (vpppad <= Vpplk) THEN                          -- vpp lock out voltage
	    e_fail := '1';
	    vpp_low := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  ELSIF (vccpad < Vlko) THEN                            -- vcc lock out voltage
	    e_fail := '1';
	    ready <= '0' AFTER tWHRL,
		     '1' AFTER tWHRL + ready_low;               -- rd_byb goes low for a short time (ready_low)
	  END IF;
	END IF;

	ELSE                                                    -- if block lock bit is set
	  e_fail := '1';                                        -- both e_fail & l_fail bits set after ???
	  l_fail := '1';
	  erasing_blk := -1;                                    -- not erasing any block (0 to 16), used negative integer
	  ready <= '0' AFTER tWHRL,                             -- rd_byb goes low for a short time (ready_low)
		   '1' AFTER tWHRL + ready_low;
	END IF;

      ELSIF (Erase_Done'EVENT) AND (Erase_Done = '1') THEN      -- if erase timer is done
	  Erase_Timer_Start <= '0';                             -- timer turned off and rd_byb goes high
	  ready <= '1';
	  erasing_blk := -1;                                    -- done erasing, variable reset
      END IF;

    WHEN EraseSuspend =>
      control_command <= Idle;                                  -- array process should do nothing
      e_susp := '1';                                            -- sometime later e_susp bit is set (???)
      Erase_Timer_Start <= '0';                                 -- timer turned off and rd_byb goes high
      ready <= '1' AFTER tWHRH2;                                -- Erase Suspend latency

    WHEN EraseSuspToArray =>
	control_command <= Read;                                -- array process should Read
      
  END CASE;                                                     --  command

  control_status <= ready & e_susp & e_fail & p_fail & vpp_low & p_susp & l_fail&"0";   -- 8-bit status register updated, to status process
  control_rdy_idle <= ready & e_susp & p_susp;                  -- suspend and ready status sent to csm process
  
END PROCESS face;

------- lock_program_timer ----------------------------------------------------

lock_program_timer : PROCESS

VARIABLE Lock_Program_Start_Time : TIME := 0 NS;
VARIABLE Lock_Program_dt         : TIME := 0 NS;
VARIABLE Lock_Program_wait       : BOOLEAN := FALSE;

BEGIN  --  PROCESS lock_program_timer 

   IF (Lock_Program_Timer_Start'EVENT) AND (Lock_Program_Timer_Start = '1') THEN        -- if timer started by FACE
      Lock_Program_Start_Time := NOW;                                                   -- save current simulation time
      Lock_Program_wait := TRUE;                                                        -- timer flag set
      Lock_Program_Done <= '0';                                                         -- set signal to indicate timer is not done
   ELSIF (Lock_Program_Timer_Start'EVENT) AND (Lock_Program_Timer_Start = '0') THEN     -- if timer is turned off by FACE
      Lock_Program_wait := FALSE;                                                       -- timer flag cleared
   END IF;

   IF (Lock_Program_dt < tWHQV3) AND (Lock_Program_wait) THEN                   -- if lock-program time not completed & timer flag set
      WAIT UNTIL ((reset'EVENT) AND (reset = '1'))                                      -- continue advancing time until reset
		 FOR (tWHQV3 - Lock_Program_dt);                                        -- or lock-program time completed
      Lock_Program_dt := NOW - Lock_Program_Start_Time + Lock_Program_dt;               -- save the time advanced
      IF (Lock_Program_dt = tWHQV3) OR (reset = '1') THEN                               -- if lock-program time completed or reset
	 Lock_Program_Done <= '1';                                                      -- set signal to indicate timer is done
	 Lock_Program_dt := 0 NS;                                                       -- reset local variable
	 Lock_Program_wait := FALSE;                                                    -- timer flag cleared
      END IF;
   END IF;

  WAIT ON Lock_Program_Timer_Start, reset;
  
END PROCESS lock_program_timer;

------- lock_erase_timer ------------------------------------------------------

lock_erase_timer : PROCESS

VARIABLE Lock_Erase_Start_Time : TIME := 0 NS;
VARIABLE Lock_Erase_dt         : TIME := 0 NS;
VARIABLE Lock_Erase_wait       : BOOLEAN := FALSE;

BEGIN  --  PROCESS lock_erase_timer 

   IF (Lock_Erase_Timer_Start'EVENT) AND (Lock_Erase_Timer_Start = '1') THEN    -- if timer started by FACE
      Lock_Erase_Start_Time := NOW;                                             -- save current simulation time
      Lock_Erase_wait := TRUE;                                                  -- timer flag set
      Lock_Erase_Done <= '0';                                                   -- set signal to indicate timer is not done
   ELSIF (Lock_Erase_Timer_Start'EVENT) AND (Lock_Erase_Timer_Start = '0') THEN -- if timer is turned off by FACE
      Lock_Erase_wait := FALSE;                                                 -- timer flag cleared
   END IF;

   IF (Lock_Erase_dt < tWHQV4) AND (Lock_Erase_wait) THEN                       -- if lock-erase time not completed & timer flag set
      WAIT UNTIL ((reset'EVENT) AND (reset = '1'))                              -- continue advancing time until reset
		 FOR (tWHQV4 - Lock_Erase_dt);                          -- or lock-erase time completed
      Lock_Erase_dt := NOW - Lock_Erase_Start_Time + Lock_Erase_dt;             -- save the time advanced
      IF (Lock_Erase_dt = tWHQV4) OR (reset = '1') THEN                 -- if lock-erase time completed or reset
	 Lock_Erase_Done <= '1';                                                -- set signal to indicate timer is done
	 Lock_Erase_dt := 0 NS;                                                 -- reset local variable
	 Lock_Erase_wait := FALSE;                                              -- timer flag cleared
      END IF;
   END IF;

  WAIT ON Lock_Erase_Timer_Start, reset;
  
END PROCESS lock_erase_timer;

------- program_timer ---------------------------------------------------------

program_timer : PROCESS

VARIABLE Program_Start_Time : TIME := 0 NS;
VARIABLE Program_dt         : TIME := 0 NS;
VARIABLE Program_wait       : BOOLEAN := FALSE;

BEGIN  --  PROCESS program_timer 

   IF (Program_Timer_Start'EVENT) AND (Program_Timer_Start = '1') THEN          -- if timer started by FACE
      Program_Start_Time := NOW;                                                -- save current simulation time
      Program_wait := TRUE;                                                     -- timer flag set
      Program_Done <= '0';                                                      -- set signal to indicate timer is not done
   ELSIF (Program_Timer_Start'EVENT) AND (Program_Timer_Start = '0') THEN       -- if timer is turned off by FACE
      Program_wait := FALSE;                                                    -- timer flag cleared
   END IF;

   IF (Program_dt < tWHQV1) AND (Program_wait) THEN                             -- if program time not completed & timer flag set
      WAIT UNTIL ( ((Program_Timer_Start'EVENT) AND (Program_Timer_Start = '0')) OR     -- continue advancing time until timer stopped
		   ((reset'EVENT) AND (reset = '1')) )                                  -- by FACE, reset or program time completed
		 FOR (tWHQV1 - Program_dt);
      Program_dt := NOW - Program_Start_Time + Program_dt;                      -- when suspended, save the time already advanced
      IF (Program_dt = tWHQV1) OR (reset = '1') THEN                            -- if program time completed or reset
	 Program_Done <= '1';                                                   -- set signal to indicate timer is done
	 Program_dt := 0 NS;                                                    -- reset local variable
	 Program_wait := FALSE;                                                 -- timer flag cleared
      END IF;
   END IF;

  WAIT ON Program_Timer_Start, reset;
  
END PROCESS program_timer;

------- erase_timer -----------------------------------------------------------

erase_timer : PROCESS

VARIABLE Erase_Start_Time : TIME := 0 NS;
VARIABLE Erase_dt         : TIME := 0 NS;
VARIABLE Erase_wait       : BOOLEAN := FALSE;

BEGIN  --  PROCESS erase_timer 

   IF (Erase_Timer_Start'EVENT) AND (Erase_Timer_Start = '1') THEN              -- if timer started by FACE
      Erase_Start_Time := NOW;                                                  -- save current simulation time
      Erase_wait := TRUE;                                                       -- timer flag set
      Erase_Done <= '0';                                                        -- set signal to indicate timer is not done
   ELSIF (Erase_Timer_Start'EVENT) AND (Erase_Timer_Start = '0') THEN           -- if timer is turned off by FACE
      Erase_wait := FALSE;                                                      -- timer flag cleared
   END IF;

   IF (Erase_dt < tWHQV2) AND (Erase_wait) THEN                                 -- if erase time not completed & timer flag set
      WAIT UNTIL ( ((Erase_Timer_Start'EVENT) AND (Erase_Timer_Start = '0')) OR         -- continue advancing time until timer stopped
		   ((reset'EVENT) AND (reset = '1')) )                                  -- by FACE, reset or erase time completed
		 FOR (tWHQV2 - Erase_dt);
      Erase_dt := NOW - Erase_Start_Time + Erase_dt;                            -- when suspended, save the time already advanced
      IF (Erase_dt = tWHQV2) OR (reset = '1') THEN                              -- if erase time completed or reset
	 Erase_Done <= '1';                                                     -- set signal to indicate timer is done
	 Erase_dt := 0 NS;                                                      -- reset local variable
	 Erase_wait := FALSE;                                                   -- timer flag cleared
      END IF;
   END IF;

  WAIT ON Erase_Timer_Start, reset;
  
END PROCESS erase_timer;


-------------------------------------------------------------------------------
--  Process:   status
--  Purpose:
--  Inputs:     addr_buf, output_en, command
--  Outputs:    read_array, read_lock, id_or_status
-------------------------------------------------------------------------------
status : PROCESS (addr_buf,output_en,command)

  VARIABLE      lat_status       : std_ulogic_vector(7 DOWNTO 0);

BEGIN  --  PROCESS status 

  IF (output_en'EVENT) AND (output_en = '1') THEN                       -- at the rising edge of output_en (OEB falling edge)
    lat_status := control_status;                                       -- latch the status output from FACE process
  END IF;

  IF (command = ReadID) THEN                                            -- output the id in this CSM state
    read_array <= '0';                                                  -- array process should not read the array when address transitions
    
    IF (addr_buf(1) = '1') THEN                                         -- if A1 = 1
      read_lock <= '1';                                                 -- read lock status from memory, A0 decoding done in array process
      id_or_status <= lat_status;                                       -- drive the status out to the output mux
    ELSIF (addr_buf(1) = '0') THEN                                      -- if A1 = 0
      read_lock <= '0';                                                 -- read vend/dev id and not the lock
      IF (addr_buf(0) = '1') THEN                                       -- if A1A0 = 01, output device id
	id_or_status <= dev_id;
      ELSIF (addr_buf(0) = '0') THEN                                    -- if A1A0 = 00, output vendor id
	id_or_status <= vend_id;
      ELSE                                                              -- if A1A0 = 0X, output is X's
	id_or_status <= "XXXXXXXX";
      END IF;
    ELSE                                                                -- if A1 = X, output is X's
      read_lock <= 'X';
      id_or_status <= "XXXXXXXX";
    END IF;
    
  ELSIF (command = ReadArray) OR (command = EraseSuspToArray) OR
    (command = ProgramSuspToArray) OR (command = ClearStatus) THEN      --  output the array in these states
    read_array<='1';                                                    -- array process should read the array when addr transitions
    read_lock <= '0';                                                   -- do not read lock status from memory
    id_or_status <= lat_status;                                         -- drive the status out to the output mux
    
  ELSE                                                                  --  in all other states output the status
    read_array <= '0';                                                  -- do not read the array
    read_lock <= '0';                                                   -- do not read the lock
    id_or_status <= lat_status;                                         -- drive the status out to the output mux
    
  END IF;
END PROCESS status;

-------------------------------------------------------------------------------
--  Process:   output_mux
--  Purpose:
--  Inputs:    mem_data_out, id_or_status, read_array, read_lock
--  Outputs:   mux_data_out
-------------------------------------------------------------------------------
output_mux : PROCESS (read_lock, read_array, id_or_status, mem_data_out)
  
BEGIN  --  PROCESS output_mux 

  IF (read_lock = '0' AND read_array = '0') THEN                        -- if 00 
    mux_data_out <= id_or_status;                                       -- output id or status
  ELSIF (read_lock = '0' AND read_array='1') THEN                       -- if 01 or 10
    mux_data_out <= mem_data_out;                                       -- output either the array or lock data from the memory
  ELSIF (read_lock = '1' AND read_array='0') THEN
    mux_data_out <= mem_data_out;
  ELSIF (read_lock = '1' AND read_array='1') THEN                       -- cannot have both the array data and lock data being read
    ASSERT (NOT verbose)                                                --     at the same time
      REPORT "output_mux: read_lock and read_array are both active."
    SEVERITY error;
  END IF;

END PROCESS output_mux;

-------------------------------------------------------------------------------
--  Output Buffer: The output of the output_mux(mux_data_out) goes through a buffer controlled
--  by output_en before being driven out to the data bus.
-------------------------------------------------------------------------------

data <= To_StdLogicVector(mux_data_out) WHEN (output_en = '1') AND (output_xs ='0')ELSE -- if output_en = '1', output the data
	"XXXXXXXX" WHEN (output_en = '0') AND (output_xs = '1') ELSE                    -- if output_xs = '1', output x's on the data bus
	"ZZZZZZZZ";                                                                     -- else drive it to Z's
	
-------------------------------------------------------------------------------
--  Ready/Busy Output: The ready signal output from FACE is driven out to the rd_byb signal. It
--  shows whether FACE is busy or not.
-------------------------------------------------------------------------------

rd_byb <= control_rdy_idle(2);

-------------------------------------------------------------------------------
--  Process:   array_proc
--  Purpose:
--  Inputs:    control_command, mem_addr, read_lock
--  Outputs:  mem_data_out
--  Limits: HV override for block locking not implemented yet.
-------------------------------------------------------------------------------
array_proc : PROCESS (control_command, mem_addr, read_lock, Program_Done, Erase_Done)

  VARIABLE      temp_addr        : bit_vector(mem_size DOWNTO 0);
  VARIABLE      temp_addr1       : bit_vector(23 DOWNTO 0);
  VARIABLE      block_addr       : integer;
  VARIABLE      link_addr        : bit_vector(15 DOWNTO 0);
  VARIABLE      link_data        : bit_vector(7 DOWNTO 0);
  VARIABLE      memory_block     : mem_blks(32 DOWNTO 0);                       --  max number of blocks (for 16M)
  VARIABLE      num_blks         : integer;                                     -- number of blocks varies with memory size (4/8/16M)

  --  variables needed to store the block number being erased or programmed.
  VARIABLE      erasing_blk           : integer := -1;                          -- initialized to something other than 32:0 block numbers
  VARIABLE      programming_blk       : integer := -1;

  --  variables needed for lock status
  VARIABLE      lock_block       : integer;                                     --  the block to lock
  VARIABLE      master_lock      : bit := '0';                                  --  master lock status
  VARIABLE      block_status     : bit_vector(31 DOWNTO 0) := X"00000000";      --  block lock status(max number of blocks is 32)

  --  variables needed for initialization.
  VARIABLE      link     : init_file_rec;
  VARIABLE      newline  : LINE;
  VARIABLE      mem_init : BOOLEAN := FALSE;

  --  variables to check for valid data and address.
  VARIABLE      addr2check       : std_ulogic_vector(mem_size DOWNTO 0);
  VARIABLE      data2check       : std_ulogic_vector(7 DOWNTO 0);
  VARIABLE      addr_valid       : BOOLEAN;
  VARIABLE      data_valid       : BOOLEAN;

  FILE          data_file: TEXT IS IN init_file;
 
BEGIN  --  PROCESS array_proc 

  IF (mem_size = 18) THEN                                                       -- depending on the memory size, this will determine the
    num_blks := 8;                                                              --    number of blocks that may exist
  ELSIF (mem_size = 19) THEN
    num_blks := 16;
  ELSIF (mem_size = 20) THEN
    num_blks := 32;
  ELSE                                                                          -- this model will only work for 4M/8M/16M memories
    ASSERT (NOT verbose)
      REPORT "array_proc: mem_size is invalid for this model."
    SEVERITY error;
  END IF;
  
-------------------------------------------------------------------------------
--  Array Initialization:Initialize the array based on the input file.
--  Input file format is hex address followed by integer data. It may also
--  contain 'lockms' to lock the master lock or 'lockbk' followed by block
--  number to lock a block.
-------------------------------------------------------------------------------
  IF (NOT(mem_init)) THEN                                                       -- if memory has not yet been initialized
    init_loop1 : FOR a IN 0 TO (num_blks - 1) LOOP
		   memory_block(a) := NULL;                                     -- initialize pointers to each block to NULL
		   block_status(a) := '0';                                      -- initialize lock status of each block to unlocked
    END LOOP init_loop1;
    READLINE(data_file,newline);                                                --  first line is left for headings
    init_loop2 : WHILE (NOT ENDFILE(data_file)) LOOP
		   READLINE(data_file,newline);
		   READ(newline,link.addr);                                     -- read first string in each line
		   
		   IF (link.addr = "lockbk") THEN                               -- if it is lock block
		     READ(newline,link.data);                                   -- read the second string to determine which block
		     lock_block:=link.data;                                     -- block number to lock in integer
		     IF (lock_block > num_blks) THEN                            -- if the block number is too high for the memory size
		       ASSERT (NOT verbose)                                     -- give an error message
			 REPORT "array_proc: Invalid block number."
		       SEVERITY error;
		     ELSE                                                       -- otherwise, set the lock bit
		       block_status(lock_block):='1';
		     END IF;
		     
		   ELSIF (link.addr = "lockms") THEN                            -- if it is to set the master lock then set the variable
		     master_lock := '1';
		     
		   ELSE                                                         -- else, it must be addr/data for programming
		     READ(newline,link.data);                                   -- read the second string also
		     temp_addr1 := HexStr_to_BitVector(link.addr);              -- convert the addr to a bit vector
		     temp_addr := temp_addr1(mem_size DOWNTO 0);                -- 'chop' it for the correct memory size
		     link_data := Int_to_BitVector(link.data, 8);               -- convert the data to a bit vector
		     block_addr := BitVector_to_Int(temp_addr(mem_size DOWNTO 16));     -- separate it into block address of type integer
		     link_addr := temp_addr(15 DOWNTO 0);                       -- & address within the block

		     program_byte(block_addr,link_addr,link_data,memory_block); -- then program that location
		     
		   END IF;
    END LOOP init_loop2;

    block_status_sig <= block_status;                                           -- status sent to FACE process
    mem_init := TRUE;                                                           -- set flag to show that memory has been initialized

  END IF;
-------------------------------------------------------------------------------

  addr2check := mem_addr;                                                       --  assigning signals to variables.
  data2check := mem_data;

  temp_addr := To_bitvector(mem_addr);                                          -- convert addr from std_ulogic_vector to bitvector
  block_addr := BitVector_to_Int(temp_addr(mem_size DOWNTO 16));                -- separate it into block address of type integer
  link_addr := temp_addr(15 DOWNTO 0);                                          -- & address within the block
  link_data := To_bitvector(mem_data);                                          -- convert data from std_ulogic_vector to bitvector

  IF (control_command'EVENT) OR (mem_addr'EVENT) OR (read_lock'EVENT) THEN
    
    IF (control_command = read) THEN
      IF (block_addr = programming_blk) OR (block_addr = erasing_blk) THEN      -- if attempting to read from a block being erased/programmed
	mem_data_out <= "XXXXXXXX";                                             -- then output X's
      ELSE
	read_byte(block_addr, link_addr, link_data, memory_block);              -- else read the byte & output data
	mem_data_out <= To_StdULogicVector(link_data);                          -- convert back to stdulogicvector from bitvector data
      END IF;

    ELSIF (control_command = program) THEN
      programming_blk := block_addr;                                            -- save the block number in which programming
      program_byte(block_addr, link_addr, link_data, memory_block);             -- program the byte

      check_data(data2check,data_valid);                                        -- check to see if data being programmed is valid
      ASSERT(data_valid AND (NOT verbose))
	REPORT "Array Process: Data Invalid."
      SEVERITY ERROR;

    ELSIF (control_command = erase) THEN
      erasing_blk := block_addr;                                                -- save the block number being erased
      erase_block(block_addr, memory_block);                                    -- erase the block

    ELSIF (control_command = lock) THEN                                         -- if doing a lock command 
      IF (link_data = X"01" AND master_lock = '0') THEN                         -- if attempting 'lock block' & master lock is not set
	block_status(block_addr) := '1';                                        -- then set the lock bit for that particular block
      ELSIF (link_data = X"D0" AND master_lock = '0') THEN                      -- if attempting 'erase all block locks' & master not set
	FOR a IN 0 TO (num_blks - 1) LOOP                                       --  erase all block locks
	  block_status(a) := '0';
	END LOOP;
      ELSIF (link_data = X"F1") THEN                                            -- if 'set master lock' command
	master_lock := '1';                                                     -- set the master lock bit
      ELSE                                                                      -- else it must be an incorrect lock command sequence
	ASSERT (NOT verbose)
	  REPORT "array_proc: Invalid Data in lock command."
	SEVERITY error;
      END IF;

      block_status_sig <= block_status;                                         -- pass the new block lock status to FACE
      
    ELSIF (control_command = idle) THEN                                         -- if FACE forces array to be 'idle'
      mem_data_out <= "ZZZZZZZZ";                                               -- data is not being driven

    END IF;

    IF (read_lock = '1') THEN
      IF (link_addr(0) = '0') THEN                                              -- if A0 = 0, read block lock status
	mem_data_out <= "0000000" & To_StdULogic(block_status(block_addr));     -- append 0's to the status bit(LSB)
      ELSIF (link_addr(0) = '1') THEN                                           -- if A0 = 1, read master lock status
	mem_data_out <= "0000000" & To_StdULogic(master_lock);                  -- append 0's to the master lock status bit(LSB)
      ELSE                                                                      -- if A0 = X, drive X's out
	mem_data_out <= "XXXXXXXX";
      END IF;
    END IF;

  END IF;

  IF (Program_Done'EVENT) AND (Program_Done = '1') THEN                         -- when programming completes
    programming_blk := -1;                                                      -- reset the block number being programmed
  ELSIF (Erase_Done'EVENT) AND (Erase_Done = '1') THEN                          -- when erasing completes
    erasing_blk := -1;                                                          -- reset the block number being erased
  END IF;

END PROCESS array_proc;


-------------------------------------------------------------------------------
--  Process:   timing_checker
--  Purpose:
--  Inputs:
--  Outputs:  
-------------------------------------------------------------------------------
timing_checker : PROCESS (rpb, ceb, web,
		    web'DELAYED(tWHEH), web'DELAYED(tWHDX), web'DELAYED(tWHAX),
		    ceb'DELAYED(tEHWH), ceb'DELAYED(tEHDX), ceb'DELAYED(tEHAX))
  VARIABLE      web_cntrl        : BOOLEAN := FALSE;
  VARIABLE      ceb_cntrl        : BOOLEAN := FALSE;

BEGIN  --  PROCESS timing_checker 
  IF (rpb = '1') OR (rpb = 'H') THEN                                                    -- RP# = 5V or 12V
											--   WEB controlled timings
    IF (web'EVENT) AND (web = '0') AND (ceb = '0') THEN                                 -- WEB falling edge
      web_cntrl := TRUE;

      IF (rpb'LAST_VALUE = '0') THEN
	assert (rpb'DELAYED'STABLE(tPHWL)) AND (NOT verbose)
	  report "Timing Checker: tPHWL time violation."
	severity ERROR;
      END IF;

      assert (ceb'DELAYED'STABLE(tELWL)) AND (NOT verbose)
	report "Timing Checker: tELWL time violation."
      severity ERROR;

    ELSIF (web'EVENT) AND (web = '1') AND (web_cntrl) THEN                              -- WEB rising edge
      web_cntrl := FALSE;

      IF (ceb = '0') THEN

	IF (rpb = '1' AND rpb'LAST_VALUE = 'H') OR (rpb = 'H' AND rpb'LAST_VALUE = '1') THEN
	assert (rpb'DELAYED'STABLE(tPHHWH)) AND (NOT verbose)
	  report "Timing Checker: tPHHWH time violation."
	severity ERROR;
	END IF;

	assert (data'DELAYED'STABLE(tDVWH)) AND (NOT verbose)
	  report "Timing Checker: tDVWH time violation."
	severity ERROR;

	assert (addr'DELAYED'STABLE(tAVWH)) AND (NOT verbose)
	  report "Timing Checker: tAVWH time violation."
	severity ERROR;

      ELSE
	assert (NOT verbose)
	  report "Timing Checker: ceb went high during a web controlled cycle."
	severity ERROR;
	
      END IF;

    ELSIF (web'DELAYED(tWHEH)'EVENT) AND (web'DELAYED(tWHEH) = '1') THEN                -- WEB rising edge delayed by tWHEH

      assert (ceb'DELAYED'STABLE(tWHEH)) AND (NOT verbose)
	report "Timing Checker: tWHEH time violation."
      severity ERROR;

    ELSIF (web'DELAYED(tWHDX)'EVENT) AND (web'DELAYED(tWHDX) = '1') THEN                -- WEB rising edge delayed by tWHDX

      assert (data'DELAYED'STABLE(tWHDX))AND (NOT verbose)
	report "Timing Checker: tWHDX time violation."
      severity ERROR;

    ELSIF (web'DELAYED(tWHAX)'EVENT) AND (web'DELAYED(tWHAX) = '1') THEN                -- WEB rising edge delayed by tWHAX

      assert (addr'DELAYED'STABLE(tWHAX)) AND (NOT verbose) 
	report "Timing Checker: tWHAX time violation."
      severity ERROR;

											--   CEB controlled timings
      
    ELSIF (ceb'EVENT) AND (ceb = '0') AND (web = '0') THEN                              -- CEB falling edge
      ceb_cntrl := TRUE;

      IF (rpb'LAST_VALUE = '0') THEN
      assert (rpb'DELAYED'STABLE(tPHEL)) AND (NOT verbose)
	report "Timing Checker: tPHEL time violation."
      severity ERROR;
      END IF;

      assert (ceb'DELAYED'STABLE(tWLEL)) AND (NOT verbose)
	report "Timing Checker: tWLEL time violation."
      severity ERROR;

    ELSIF (ceb'EVENT) AND (ceb = '1') AND (ceb_cntrl) THEN                              -- CEB rising edge
      ceb_cntrl := FALSE;

      IF (web = '0') THEN

	IF (rpb = '1' AND rpb'LAST_VALUE = 'H') OR (rpb = 'H' AND rpb'LAST_VALUE = '1') THEN
	assert (rpb'DELAYED'STABLE(tPHHEH)) AND (NOT verbose)
	  report "Timing Checker: tPHHEH time violation."
	severity ERROR;
	END IF;

	assert (data'DELAYED'STABLE(tDVEH)) AND (NOT verbose)
	  report "Timing Checker: tDVEH time violation."
	severity ERROR;

	assert (addr'DELAYED'STABLE(tAVEH)) AND (NOT verbose)
	  report "Timing Checker: tAVEH time violation."
	severity ERROR;

      ELSE
	assert (NOT verbose)
	  report "Timing Checker: web went high during a ceb controlled cycle."
	severity ERROR;
	
      END IF;

    ELSIF (ceb'DELAYED(tEHWH)'EVENT) AND (ceb'DELAYED(tEHWH) = '1') THEN                -- CEB rising edge delayed by tEHWH

      assert (web'DELAYED'STABLE(tEHWH)) AND (NOT verbose)
	report "Timing Checker: tEHWH time violation."
      severity ERROR;

    ELSIF (ceb'DELAYED(tEHDX)'EVENT) AND (ceb'DELAYED(tEHDX) = '1') THEN                -- CEB rising edge delayed by tEHDX

      assert (data'DELAYED'STABLE(tEHDX))AND (NOT verbose)
	report "Timing Checker: tEHDX time violation."
      severity ERROR;

    ELSIF (ceb'DELAYED(tEHAX)'EVENT) AND (ceb'DELAYED(tEHAX) = '1') THEN                -- CEB rising edge delayed by tEHAX

      assert (addr'DELAYED'STABLE(tEHAX)) AND (NOT verbose) 
	report "Timing Checker: tEHAX time violation."
      severity ERROR;

    END IF;


  ELSIF (rpb = '0') THEN                                                                -- RP# = 0V
    --  then what??
    web_cntrl := FALSE;
    ceb_cntrl := FALSE;
    
  END IF;
END PROCESS timing_checker;

END rev1;

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
